#include <Winsock2.h>
#include <Windows.h>
#include "scripts.h"
#include "engine.h"
#include "Misc.h"
extern "C" 
{
	#include "LuaLib\lua.h"
	#include "LuaLib\lualib.h"
	#include "LuaLib\lauxlib.h"
}
#include "LuaManager.h"
#include "LuaFunctions.h"
#include "LuaHooks.h"
#include "Lua_Class.h"

#pragma warning(disable: 4291)

LuaScriptManager::Script *LuaScriptManager::Scripts[256];

void LuaScriptManager::Load()
{
	memset((void *)Scripts, 0, 256*4);
}

ScriptImpClass *LuaScriptManager::Get_Script(GameObject *obj, int scriptID)
{
	SimpleDynVecClass<GameObjObserverClass *> *observers = &((ScriptableGameObj *)obj)->Observers;
	int x = observers->Count();
	for (int i = 0;i < x;i++)
	{
		if((*observers)[i]->GetID() == scriptID)
		{
			return ((ScriptImpClass*)(*observers)[i]);
		}
	}
	return 0;
}

void LuaScriptManager::Lua_DestroyScript(GameObject *obj, int scriptID)
{
	if(!obj)
	{
		return;
	}
	ScriptImpClass *script = Get_Script(obj, scriptID);
	if(!script)
	{
		return;
	}
	script->Destroy_Script();
}

float LuaScriptManager::Lua_Get_Float_Parameter(GameObject *obj, int scriptID, const char *Name)
{
	if(!obj)
	{
		return 0;
	}
	ScriptImpClass *script = Get_Script(obj, scriptID);
	if(!script)
	{
		return 0;
	}
	return script->Get_Float_Parameter(Name);
}

int LuaScriptManager::Lua_Get_Int_Parameter(GameObject *obj, int scriptID, const char *Name)
{
	if(!obj)
	{
		return 0;
	}
	ScriptImpClass *script = Get_Script(obj, scriptID);
	if(!script)
	{
		return 0;
	}
	return script->Get_Int_Parameter(Name);
}

const char *LuaScriptManager::Lua_Get_String_Parameter(GameObject *obj, int scriptID, const char *Name)
{
	if(!obj)
	{
		return "";
	}
	ScriptImpClass *script = Get_Script(obj, scriptID);
	if(!script)
	{
		return "";
	}
	return script->Get_Parameter(Name);
}

void LuaScriptManager::AddScript(const char *LuaFile)
{
	lua_State *L = lua_open();
	luaL_openlibs(L);

	LuaSLNodeWrapper::Register(L);
	AddFunctions(L);
	RegisterScriptFunctions(L);
	LuaSocketManager::Register(L);
	int s = luaL_loadfile(L, LuaFile);
	if (s == 0) 
	{
		s = lua_pcall(L, 0, LUA_MULTRET, 0);
	}
	if(Report_Errors(L, s))
	{
		lua_close(L);
		return;
	}
	const char *Name, *Params;
	
	lua_getglobal(L, "ScriptName");
	if(lua_isfunction(L, -1))
	{
		if(!LuaManager::Report_Errors(L, lua_pcall(L, 0, 1, 0)))
		{
			Name = lua_tostring(L, -1);
			lua_pop(L, 1);
		}
		else
		{
			lua_close(L);
			return;
		}
	}
	else
	{
		printf("Failed to load script inside %s: could not find function \"ScriptParams\"\n", LuaFile);
		lua_close(L);
		return;
	}
		
	for(int script_it = 0; script_it < 256; script_it++)
	{
		Script *s = Scripts[script_it];
		if(!s)
		{
			continue;
		}
		if(strcmp(s->script, Name) == 0)
		{
			printf("Found duplicate script: \"%s\", not adding script.", Name);
			lua_close(L);
			return;
		}
	}
	


	lua_getglobal(L, "ScriptParams");
	if(lua_isfunction(L, -1))
	{
		if(!LuaManager::Report_Errors(L, lua_pcall(L, 0, 1, 0)))
		{
			Params = lua_tostring(L, -1);
			lua_pop(L, 1);
		}
		else
		{
			lua_close(L);
			return;
		}
	}
	else
	{
		printf("Failed to load script inside %s: could not find function \"ScriptParams\"\n", LuaFile);
		lua_close(L);
		return;
	}

	Script *script = new Script;
	ScriptRegistrant<LuaScript> *reg = new ScriptRegistrant<LuaScript>((char *)Name, (char *)Params);
	script->script = reg->GetName();
	script->lua = L;
	script->scriptreg = reg;
	for(int script_it = 0; script_it < 256; script_it++)
	{
		Script *s = Scripts[script_it];
		if(!s)
		{
			Scripts[script_it] = script;
			break;
		}

	}
	printf("Loaded script: %s\n", Name);
	
}

bool LuaScriptManager::Report_Errors(lua_State *L, int status)
{
	if(status != 0) 
	{
		printf("%s\n", lua_tostring(L, -1));
		lua_pop(L, 1); 
		return 1;
	}
	return 0;
}

void LuaScriptManager::Lua_Start_Timer(GameObject *obj, int scriptID, float Time, int Number)
{
	if(!obj)
	{
		return;
	}
	ScriptImpClass *script = Get_Script(obj, scriptID);
	if(!script)
	{
		return;
	}
	Commands->Start_Timer(obj, script, Time, Number);
}



void LuaScriptManager::Cleanup()
{
	for(int script_it = 0; script_it < 256; script_it++)
	{
		Script *s = Scripts[script_it];
		if(!s)
		{
			continue;
		}
		lua_close(s->lua);
		delete s->scriptreg;
		delete s;
		
	}
	memset((void *)Scripts, 0, 256*4);
}

void LuaScriptManager::RegisterScriptFunctions(lua_State *L)
{
	lua_register(L, "DestroyScript", LuaScriptManager::Lua_DestroyScript_Wrap);
	lua_register(L, "Get_Float_Parameter", LuaScriptManager::Lua_Get_Float_Parameter_Wrap);
	lua_register(L, "Get_Int_Parameter", LuaScriptManager::Lua_Get_Int_Parameter_Wrap);
	lua_register(L, "Get_String_Parameter", LuaScriptManager::Lua_Get_String_Parameter_Wrap);
	lua_register(L, "Start_Timer", LuaScriptManager::Lua_Start_Timer_Wrap);
}
int LuaScriptManager::Lua_DestroyScript_Wrap(lua_State *L)
{
	int argc = lua_gettop(L);
	if(argc < 2)
	{
		return 0;
	}
	int scriptID = (int)lua_tonumber(L, 1);
	int objID = (int)lua_tonumber(L, 2);
	LuaScriptManager::Lua_DestroyScript(Commands->Find_Object(objID), scriptID);
	return 0;
}
int LuaScriptManager::Lua_Get_Float_Parameter_Wrap(lua_State *L)
{
	int argc = lua_gettop(L);
	if(argc < 3)
	{
		return 0;
	}
	int scriptID = (int)lua_tonumber(L, 1);
	int objID = (int)lua_tonumber(L, 2);
	const char *name = lua_tostring(L, 3);
	lua_pushnumber(L, LuaScriptManager::Lua_Get_Float_Parameter(Commands->Find_Object(objID), scriptID, name));
	return 1;
}
int LuaScriptManager::Lua_Get_Int_Parameter_Wrap(lua_State *L)
{
	int argc = lua_gettop(L);
	if(argc < 3)
	{
		return 0;
	}
	int scriptID = (int)lua_tonumber(L, 1);
	int objID = (int)lua_tonumber(L, 2);
	const char *name = lua_tostring(L, 3);
	lua_pushnumber(L, LuaScriptManager::Lua_Get_Int_Parameter(Commands->Find_Object(objID), scriptID, name));
	return 1;
}
int LuaScriptManager::Lua_Get_String_Parameter_Wrap(lua_State *L)
{
	int argc = lua_gettop(L);
	if(argc < 3)
	{
		return 0;
	}
	int scriptID = (int)lua_tonumber(L, 1);
	int objID = (int)lua_tonumber(L, 2);
	const char *name = lua_tostring(L, 3);
	lua_pushstring(L, LuaScriptManager::Lua_Get_String_Parameter(Commands->Find_Object(objID), scriptID, name));
	return 1;
}
int LuaScriptManager::Lua_Start_Timer_Wrap(lua_State *L)
{
	int argc = lua_gettop(L);
	if(argc < 4)
	{
		return 0;
	}
	int scriptID = (int)lua_tonumber(L, 1);
	int obj = (int)lua_tonumber(L, 2);
	float Time = (float)lua_tonumber(L, 3);
	int Number = (int)lua_tonumber(L, 4);
	LuaScriptManager::Lua_Start_Timer(Commands->Find_Object(obj), scriptID, Time, Number);
	return 0;
}

//The dynamic script itself
void LuaScriptManager::LuaScript::Created(GameObject *obj)
{
	for(int script_it = 0; script_it < 256; script_it++)
	{
		Script *s = Scripts[script_it];
		if(!s)
		{
			continue;
		}
		if(strcmp(s->script, this->Get_Name()) == 0)
		{
			lua_State *L = s->lua;
			if(!L)
			{
				continue;
			}
			lua_getglobal(L, "Created");
			if(lua_isfunction(L, -1))
			{
				lua_pushnumber(L, GetID());
				lua_pushnumber(L, Commands->Get_ID(obj));
				LuaManager::Report_Errors(L, lua_pcall(L, 2, 0, 0));
			}
			else
			{
				lua_setglobal(L, "Created");
			}
		}
	}
}
void LuaScriptManager::LuaScript::Destroyed(GameObject *obj)
{
	for(int script_it = 0; script_it < 256; script_it++)
	{
		Script *s = Scripts[script_it];
		if(!s)
		{
			continue;
		}
		if(strcmp(s->script, this->Get_Name()) == 0)
		{
			lua_State *L = s->lua;
			if(!L)
			{
				continue;
			}
			lua_getglobal(L, "Destroyed");
			if(lua_isfunction(L, -1))
			{
				lua_pushnumber(L, GetID());
				lua_pushnumber(L, Commands->Get_ID(obj));
				LuaManager::Report_Errors(L, lua_pcall(L, 2, 0, 0));
			}
			else
			{
				lua_setglobal(L, "Destroyed");
			}
		}
	}
}
void LuaScriptManager::LuaScript::Killed(GameObject *obj,GameObject *shooter)
{
	for(int script_it = 0; script_it < 256; script_it++)
	{
		Script *s = Scripts[script_it];
		if(!s)
		{
			continue;
		}
		if(strcmp(s->script, this->Get_Name()) == 0)
		{
			lua_State *L = s->lua;
			if(!L)
			{
				continue;
			}
			lua_getglobal(L, "Killed");
			if(lua_isfunction(L, -1))
			{
				lua_pushnumber(L, GetID());
				lua_pushnumber(L, Commands->Get_ID(obj));
				lua_pushnumber(L, Commands->Get_ID(shooter));
				LuaManager::Report_Errors(L, lua_pcall(L, 3, 0, 0));
			}
			else
			{
				lua_setglobal(L, "Killed");
			}
		}
	}
}
void LuaScriptManager::LuaScript::Damaged(GameObject *obj,GameObject *damager,float damage)
{
	for(int script_it = 0; script_it < 256; script_it++)
	{
		Script *s = Scripts[script_it];
		if(!s)
		{
			continue;
		}
		if(strcmp(s->script, this->Get_Name()) == 0)
		{
			lua_State *L = s->lua;
			if(!L)
			{
				continue;
			}
			lua_getglobal(L, "Damaged");
			if(lua_isfunction(L, -1))
			{
				lua_pushnumber(L, GetID());
				lua_pushnumber(L, Commands->Get_ID(obj));
				lua_pushnumber(L, Commands->Get_ID(damager));
				lua_pushnumber(L, damage);
				LuaManager::Report_Errors(L, lua_pcall(L, 4, 0, 0));
			}
			else
			{
				lua_setglobal(L, "Damaged");
			}
		}
	}
}
void LuaScriptManager::LuaScript::Custom(GameObject *obj,int message,int param,GameObject *sender)
{
	for(int script_it = 0; script_it < 256; script_it++)
	{
		Script *s = Scripts[script_it];
		if(!s)
		{
			continue;
		}
		if(strcmp(s->script, this->Get_Name()) == 0)
		{
			lua_State *L = s->lua;
			if(!L)
			{
				continue;
			}
			lua_getglobal(L, "Custom");
			if(lua_isfunction(L, -1))
			{
				lua_pushnumber(L, GetID());
				lua_pushnumber(L, Commands->Get_ID(obj));
				lua_pushnumber(L, message);
				lua_pushnumber(L, param);
				lua_pushnumber(L, Commands->Get_ID(sender));
				LuaManager::Report_Errors(L, lua_pcall(L, 5, 0, 0));
			}
			else
			{
				lua_setglobal(L, "Custom");
			}
		}
	}
}

void LuaScriptManager::LuaScript::Enemy_Seen(GameObject *obj,GameObject *seen)
{
	for(int script_it = 0; script_it < 256; script_it++)
	{
		Script *s = Scripts[script_it];
		if(!s)
		{
			continue;
		}
		if(strcmp(s->script, this->Get_Name()) == 0)
		{
			lua_State *L = s->lua;
			if(!L)
			{
				continue;
			}
			lua_getglobal(L, "Enemy_Seen");
			if(lua_isfunction(L, -1))
			{
				lua_pushnumber(L, GetID());
				lua_pushnumber(L, Commands->Get_ID(obj));
				lua_pushnumber(L, Commands->Get_ID(seen));
				LuaManager::Report_Errors(L, lua_pcall(L, 3, 0, 0));
			}
			else
			{
				lua_setglobal(L, "Enemy_Seen");
			}
		}
	}
}
void LuaScriptManager::LuaScript::Action_Complete(GameObject *obj,int action, ActionCompleteReason reason)
{
	for(int script_it = 0; script_it < 256; script_it++)
	{
		Script *s = Scripts[script_it];
		if(!s)
		{
			continue;
		}
		if(strcmp(s->script, this->Get_Name()) == 0)
		{
			lua_State *L = s->lua;
			if(!L)
			{
				continue;
			}
			lua_getglobal(L, "Action_Complete");
			if(lua_isfunction(L, -1))
			{
				lua_pushnumber(L, GetID());
				lua_pushnumber(L, Commands->Get_ID(obj));
				lua_pushnumber(L, action);

				LuaManager::Report_Errors(L, lua_pcall(L, 3, 0, 0));
			}
			else
			{
				lua_setglobal(L, "Action_Complete");
			}
		}
	}
}
void LuaScriptManager::LuaScript::Timer_Expired(GameObject *obj,int number)
{
	for(int script_it = 0; script_it < 256; script_it++)
	{
		Script *s = Scripts[script_it];
		if(!s)
		{
			continue;
		}
		if(strcmp(s->script, this->Get_Name()) == 0)
		{
			lua_State *L = s->lua;
			if(!L)
			{
				continue;
			}
			lua_getglobal(L, "Timer_Expired");
			if(lua_isfunction(L, -1))
			{
				lua_pushnumber(L, GetID());
				lua_pushnumber(L, Commands->Get_ID(obj));
				lua_pushnumber(L, number);
				LuaManager::Report_Errors(L, lua_pcall(L, 3, 0, 0));
			}
			else
			{
				lua_setglobal(L, "Timer_Expired");
			}
		}
	}
}
void LuaScriptManager::LuaScript::Animation_Complete(GameObject *obj,const char *anim)
{
	for(int script_it = 0; script_it < 256; script_it++)
	{
		Script *s = Scripts[script_it];
		if(!s)
		{
			continue;
		}
		if(strcmp(s->script, this->Get_Name()) == 0)
		{
			lua_State *L = s->lua;
			if(!L)
			{
				continue;
			}
			lua_getglobal(L, "Animation_Complete");
			if(lua_isfunction(L, -1))
			{
				lua_pushnumber(L, GetID());
				lua_pushnumber(L, Commands->Get_ID(obj));
				lua_pushstring(L, anim);
				LuaManager::Report_Errors(L, lua_pcall(L, 3, 0, 0));
			}
			else
			{
				lua_setglobal(L, "Animation_Complete");
			}
		}
	}
}
void LuaScriptManager::LuaScript::Poked(GameObject *obj,GameObject *poker)
{
	for(int script_it = 0; script_it < 256; script_it++)
	{
		Script *s = Scripts[script_it];
		if(!s)
		{
			continue;
		}
		if(strcmp(s->script, this->Get_Name()) == 0)
		{
			lua_State *L = s->lua;
			if(!L)
			{
				continue;
			}
			lua_getglobal(L, "Poked");
			if(lua_isfunction(L, -1))
			{
				lua_pushnumber(L, GetID());
				lua_pushnumber(L, Commands->Get_ID(obj));
				lua_pushnumber(L, Commands->Get_ID(poker));
				LuaManager::Report_Errors(L, lua_pcall(L, 3, 0, 0));
			}
			else
			{
				lua_setglobal(L, "Poked");
			}
		}
	}
}
void LuaScriptManager::LuaScript::Entered(GameObject *obj,GameObject *enter)
{
	for(int script_it = 0; script_it < 256; script_it++)
	{
		Script *s = Scripts[script_it];
		if(!s)
		{
			continue;
		}
		if(strcmp(s->script, this->Get_Name()) == 0)
		{
			lua_State *L = s->lua;
			if(!L)
			{
				continue;
			}
			lua_getglobal(L, "Entered");
			if(lua_isfunction(L, -1))
			{
				lua_pushnumber(L, GetID());
				lua_pushnumber(L, Commands->Get_ID(obj));
				lua_pushnumber(L, Commands->Get_ID(enter));
				LuaManager::Report_Errors(L, lua_pcall(L, 3, 0, 0));
			}
			else
			{
				lua_setglobal(L, "Entered");
			}
		}
	}
}
void LuaScriptManager::LuaScript::Exited(GameObject *obj,GameObject *exit)
{
	for(int script_it = 0; script_it < 256; script_it++)
	{
		Script *s = Scripts[script_it];
		if(!s)
		{
			continue;
		}
		if(strcmp(s->script, this->Get_Name()) == 0)
		{
			lua_State *L = s->lua;
			if(!L)
			{
				continue;
			}
			lua_getglobal(L, "Exited");
			if(lua_isfunction(L, -1))
			{
				lua_pushnumber(L, GetID());
				lua_pushnumber(L, Commands->Get_ID(obj));
				lua_pushnumber(L, Commands->Get_ID(exit));
				LuaManager::Report_Errors(L, lua_pcall(L, 3, 0, 0));
			}
			else
			{
				lua_setglobal(L, "Exited");
			}
		}
	}
}



//LuaManager

lua_State *LuaManager::Lua[256];

void LuaManager::Load()
{
	memset((void *)Lua, 0, 256*4);
	
	int Loc = (int)main_loop_glue - 0x0043BAC8 - 5;
	char h[5];
	*h = '\xE9';
	memcpy((void *)(h+1), (void *)&Loc, 4);

	unsigned long op;
	void *p = OpenProcess(PROCESS_ALL_ACCESS, 0, GetCurrentProcessId());
	VirtualProtectEx(p, (void *)0x0043BAC8, 5, PAGE_EXECUTE_READWRITE, &op);
	WriteProcessMemory(p, (void *)0x0043BAC8, h, 5, NULL);
	VirtualProtectEx(p, (void *)0x0043BAC8, 5, op, NULL);
}

void LuaManager::ShowBanner()
{
	printf("|---------------------------------------------------|\n");
	printf("| Lua 5.0 Copyright (C) 1994-2003 Tecgraf, PUC-Rio  |\n");
	printf("| R. Ierusalimschy, L. H. de Figueiredo & W. Celes  |\n");
	printf("| www.lua.org                                       |\n");
	printf("| Embedded with scripts.dll by RoShamBo             |\n");
	printf("| www.dcomproductions.net                           |\n");
	printf("|---------------------------------------------------|\n");
}

void LuaManager::LoadLua(const char *LuaFile)
{
	lua_State *L = lua_open();
	luaL_openlibs(L);
	
	LuaSocketManager::Register(L);
	LuaSLNodeWrapper::Register(L);
	AddFunctions(L);
	int s = luaL_loadfile(L, LuaFile);
	if (s == 0) 
	{
		s = lua_pcall(L, 0, LUA_MULTRET, 0);
	}
	if(Report_Errors(L, s))
	{
		lua_close(L);
		printf("Failed to load plugin: %s\n", LuaFile);
		return;
	}
	printf("Loaded plugin: %s\n", LuaFile);
	
	for(int lua_it = 0; lua_it < 256; lua_it++)
	{
		if(!Lua[lua_it])
		{
			Lua[lua_it] = L;
			break;
		}
	}
}

void LuaManager::Cleanup()
{
	for(int lua_it = 0; lua_it < 256; lua_it++)
	{
		if(Lua[lua_it])
		{
			lua_close(Lua[lua_it]);
			Lua[lua_it] = 0;
		}
	}
}

bool LuaManager::Report_Errors(lua_State *L, int status)
{
	if(status != 0) 
	{
		printf("%s\n", lua_tostring(L, -1));
		lua_pop(L, 1); 
		return 1;
	}
	return 0;
}

//hooks

void LuaManager::Call_Chat_Hook(int ID, int Type, const wchar_t *Msg)
{
	for(int lua_it = 0; lua_it < 256; lua_it++)
	{
		lua_State *L = Lua[lua_it];
		if(!L)
		{
			continue;
		}
		lua_getglobal(L, "OnChat");
		if(lua_isfunction(L, -1))
		{
			const char *Msg2 = WideCharToChar(Msg);
			lua_pushnumber(L, ID);
			lua_pushnumber(L, Type);
			lua_pushstring(L, Msg2);
			LuaManager::Report_Errors(L, lua_pcall(L, 3, 0, 0));
			delete []Msg2;
		}
		else
		{
			lua_setglobal(L, "OnChat");
		}
	}
}
void LuaManager::Call_Host_Hook(int ID, int Type, const char *Msg)
{
	for(int lua_it = 0; lua_it < 256; lua_it++)
	{
		lua_State *L = Lua[lua_it];
		if(!L)
		{
			continue;
		}
		lua_getglobal(L, "OnHostMessage");
		if(lua_isfunction(L, -1))
		{
			lua_pushnumber(L, ID);
			lua_pushnumber(L, Type);
			lua_pushstring(L, Msg);
			LuaManager::Report_Errors(L, lua_pcall(L, 3, 0, 0));
		}
		else
		{
			lua_setglobal(L, "OnHostMessage");
		}
	}
}
void LuaManager::Call_Player_Join_Hook(int ID, const char *Nick)
{
	for(int lua_it = 0; lua_it < 256; lua_it++)
	{
		lua_State *L = Lua[lua_it];
		if(!L)
		{
			continue;
		}
		lua_getglobal(L, "OnPlayerJoin");
		if(lua_isfunction(L, -1))
		{
			lua_pushnumber(L, ID);
			lua_pushstring(L, Nick);
			LuaManager::Report_Errors(L, lua_pcall(L, 2, 0, 0));
		}
		else
		{
			lua_setglobal(L, "OnPlayerJoin");
		}
	}
}
void LuaManager::Call_Player_Leave_Hook(int ID)
{
	for(int lua_it = 0; lua_it < 256; lua_it++)
	{
		lua_State *L = Lua[lua_it];
		if(!L)
		{
			continue;
		}
		
		lua_getglobal(L, "OnPlayerLeave");
		if(lua_isfunction(L, -1))
		{
			lua_pushnumber(L, ID);
			LuaManager::Report_Errors(L, lua_pcall(L, 1, 0, 0));
		}
		else
		{
			lua_setglobal(L, "OnPlayerLeave");
		}
	}
}
void LuaManager::Call_Level_Loaded_Hook()
{
	for(int lua_it = 0; lua_it < 256; lua_it++)
	{
		lua_State *L = Lua[lua_it];
		if(!L)
		{
			continue;
		}
		
		lua_getglobal(L, "OnLevelLoaded");
		if(lua_isfunction(L, -1))
		{
			LuaManager::Report_Errors(L, lua_pcall(L, 0, 0, 0));
		}
		else
		{
			lua_setglobal(L, "OnLevelLoaded");
		}
	}
}
void LuaManager::Call_GameOver_Hook()
{
	for(int lua_it = 0; lua_it < 256; lua_it++)
	{
		lua_State *L = Lua[lua_it];
		if(!L)
		{
			continue;
		}
		
		lua_getglobal(L, "OnGameOver");
		if(lua_isfunction(L, -1))
		{
			LuaManager::Report_Errors(L, lua_pcall(L, 0, 0, 0));
		}
		else
		{
			lua_setglobal(L, "OnGameOver");
		}
	}
}
void LuaManager::Call_Console_Output_Hook(const char *Output)
{
	for(int lua_it = 0; lua_it < 256; lua_it++)
	{
		lua_State *L = Lua[lua_it];
		if(!L)
		{
			continue;
		}
		
		lua_getglobal(L, "OnConsoleOutput");
		if(lua_isfunction(L, -1))
		{
			lua_pushstring(L, Output);
			LuaManager::Report_Errors(L, lua_pcall(L, 1, 0, 0));
		}
		else
		{
			lua_setglobal(L, "OnConsoleOutput");
		}
	}
}
void LuaManager::Call_Version_Hook(int ID, float Ver)
{
	for(int lua_it = 0; lua_it < 256; lua_it++)
	{
		lua_State *L = Lua[lua_it];
		if(!L)
		{
			continue;
		}
		
		lua_getglobal(L, "OnPlayerVersion");
		if(lua_isfunction(L, -1))
		{
			lua_pushnumber(L, ID);
			lua_pushnumber(L, Ver);
			LuaManager::Report_Errors(L, lua_pcall(L, 2, 0, 0));
		}
		else
		{
			lua_setglobal(L, "OnPlayerVersion");
		}	
	}
}

void LuaManager::Purchase_Hook(BaseControllerClass *base, GameObject *purchaser, unsigned int cost, unsigned int preset,unsigned int purchaseret,const char *data)
{
	for(int lua_it = 0; lua_it < 256; lua_it++)
	{
		lua_State *L = Lua[lua_it];
		if(!L)
		{
			continue;
		}
		
		lua_getglobal(L, data);
		if(lua_isfunction(L, -1))
		{
			lua_pushnumber(L, Commands->Get_ID(purchaser));
			lua_pushnumber(L, cost);
			lua_pushnumber(L, preset);
			lua_pushnumber(L, purchaseret);
			LuaManager::Report_Errors(L, lua_pcall(L, 1, 0, 0));
		}
		else
		{
			lua_setglobal(L, data);
		}
	}
}

void LuaManager::Call_Object_Hook(void *data, GameObject *obj)
{
	for(int lua_it = 0; lua_it < 256; lua_it++)
	{
		lua_State *L = Lua[lua_it];
		if(!L)
		{
			continue;
		}
		lua_getglobal(L, "OnObjectCreate");
		if(lua_isfunction(L, -1))
		{
			lua_pushnumber(L, Commands->Get_ID(obj));
			LuaManager::Report_Errors(L, lua_pcall(L, 1, 0, 0));
		}
		else
		{
			lua_setglobal(L, "OnObjectCreate");
		}
	}
}

void LuaManager::Call_DDE_Hook(const char *DDEStr)
{
	for(int lua_it = 0; lua_it < 256; lua_it++)
	{
		lua_State *L = Lua[lua_it];
		if(!L)
		{
			continue;
		}
		lua_getglobal(L, "OnDDE");
		if(lua_isfunction(L, -1))
		{
			lua_pushstring(L, DDEStr);
			LuaManager::Report_Errors(L, lua_pcall(L, 1, 0, 0));
		}
		else
		{
			lua_setglobal(L, "OnDDE");
		}
	}
}

void LuaManager::Call_Think_Hook()
{
	for(int lua_it = 0; lua_it < 256; lua_it++)
	{
		lua_State *L = Lua[lua_it];
		if(!L)
		{
			continue;
		}
		lua_getglobal(L, "OnThink");
		if(lua_isfunction(L, -1))
		{
			LuaManager::Report_Errors(L, lua_pcall(L, 0, 0, 0));
		}
		else
		{
			lua_setglobal(L, "OnThink");
		}
	}
}

void LuaManager::CallInvoke(const char *Function, const char *Arg)
{
	for(int lua_it = 0; lua_it < 256; lua_it++)
	{
		lua_State *L = Lua[lua_it];
		if(!L)
		{
			continue;
		}
		lua_getglobal(L, Function);
		if(lua_isfunction(L, -1))
		{
			lua_pushstring(L, Arg);
			LuaManager::Report_Errors(L, lua_pcall(L, 1, 0, 0));
		}
	}
}

//socket manager

TCPSocket *LuaSocketManager::Sockets[64];

void LuaSocketManager::Load()
{
	memset((void *)Sockets, 0, 4*64);
	WSADATA wsaData;
	if(WSAStartup(0x202, &wsaData))
	{
		printf("Error starting winsock, LuaSocket may not work.\n");
	}

}

int LuaSocketManager::Connect(const char *IP, int port)
{
	TCPSocket *s = new TCPSocket;

	if(!s->Client(IP, port))
	{
		delete s;
		return -1;
	}

	for(int i = 0; i < 64; i++)
	{
		if(!Sockets[i])
		{
			Sockets[i] = s;
			return i;
		}
	}
	delete s;
	return -1;
}

int LuaSocketManager::Server(const char *IP, int port)
{
	TCPSocket *s = new TCPSocket;

	if(!s->Server(IP, port, 10))
	{
		delete s;
		return -1;
	}

	for(int i = 0; i < 64; i++)
	{
		if(!Sockets[i])
		{
			Sockets[i] = s;
			return i;
		}
	}
	delete s;
	return -1;
}

int LuaSocketManager::Accept(int Socket)
{
	if(Socket < 0)
	{
		return -1;
	}
	TCPSocket *s = Sockets[Socket];
	if(!s)
	{
		return -1;
	}

	TCPSocket *c = new TCPSocket;
	if(!s->Accept(c))
	{
		delete c;
		return -1;
	}
	
	for(int i = 0; i < 64; i++)
	{
		if(!Sockets[i])
		{
			Sockets[i] = c;
			return i;
		}
	}
	delete c;
	return -1;
}

int LuaSocketManager::Recv(int Socket, char *Data, int Length)
{
	if(Socket < 0)
	{
		return -1;
	}
	TCPSocket *s = Sockets[Socket];
	if(!s)
	{
		return -1;
	}

	if(!s->RecviveData(Data, &Length))
	{
		return -1;
	}
	Data[Length] = 0;
	return 0;
}

int LuaSocketManager::Send(int Socket, const char *Data)
{
	if(Socket < 0)
	{
		return -1;
	}
	TCPSocket *s = Sockets[Socket];
	if(!s)
	{
		return -1;
	}

	if(!s->SendData(Data, strlen(Data)))
	{
		return -1;
	}
	return 0;
}

int LuaSocketManager::Is_DataAvaliable(int Socket)
{
	if(Socket < 0)
	{
		return -1;
	}
	TCPSocket *s = Sockets[Socket];
	if(!s)
	{
		return -1;
	}

	if(!s->Is_DataAvaliable())
	{
		return -1;
	}
	return 0;
}

int LuaSocketManager::Is_ClientWaiting(int Socket)
{
	if(Socket < 0)
	{
		return -1;
	}
	TCPSocket *s = Sockets[Socket];
	if(!s)
	{
		return -1;
	}

	if(!s->Is_ConnectionWaiting())
	{
		return -1;
	}
	return 0;
}

void LuaSocketManager::Disconnect(int Socket)
{
	if(Socket < 0)
	{
		return;
	}
	TCPSocket *s = Sockets[Socket];
	if(!s)
	{
		return;
	}

	s->Destroy();
	delete s;
	Sockets[Socket] = 0;
}


int LuaSocketManager::Lua_Connect_Wrap(lua_State *L)
{

	int argc = lua_gettop(L);
	if(argc < 2)
	{
		return 0;
	}
	const char *Host = lua_tostring(L, 1);
	int Port = (int)lua_tonumber(L, 2);
	lua_pushnumber(L, Connect(Host, Port));
	return 1;
}

int LuaSocketManager::Lua_Server_Wrap(lua_State *L)
{
	int argc = lua_gettop(L);
	if(argc < 2)
	{
		return 0;
	}
	const char *Host = lua_tostring(L, 1);
	int Port = (int)lua_tonumber(L, 2);
	lua_pushnumber(L, Server(Host, Port));
	return 1;
}

int LuaSocketManager::Lua_Accept_Wrap(lua_State *L)
{
	int argc = lua_gettop(L);
	if(argc < 1)
	{
		return 0;
	}
	int Sock = (int)lua_tonumber(L, 1);
	lua_pushnumber(L, Accept(Sock));
	return 1;
}

int LuaSocketManager::Lua_Recv_Wrap(lua_State *L)
{
	int argc = lua_gettop(L);
	if(argc < 1)
	{
		return 0;
	}
	char data[1024];
	*data = 0;
	int Sock = (int)lua_tonumber(L, 1);
	Recv(Sock, data, 1023);
	lua_pushstring(L, data);
	return 1;
}

int LuaSocketManager::Lua_Send_Wrap(lua_State *L)
{
	int argc = lua_gettop(L);
	if(argc < 2)
	{
		return 0;
	}
	int Sock = (int)lua_tonumber(L, 1);
	const char *str = lua_tostring(L, 2);
	lua_pushnumber(L, Send(Sock, str));
	return 1;
}

int LuaSocketManager::Lua_DataAvaliable_Wrap(lua_State *L)
{
	int argc = lua_gettop(L);
	if(argc < 1)
	{
		return 0;
	}
	int Sock = (int)lua_tonumber(L, 1);
	lua_pushnumber(L, Is_DataAvaliable(Sock));
	return 1;
}

int LuaSocketManager::Lua_ClientWaiting_Wrap(lua_State *L)
{
	int argc = lua_gettop(L);
	if(argc < 1)
	{
		return 0;
	}
	int Sock = (int)lua_tonumber(L, 1);
	lua_pushnumber(L, Is_ClientWaiting(Sock));
	return 1;
}

int LuaSocketManager::Lua_Disconnect_Wrap(lua_State *L)
{
	int argc = lua_gettop(L);
	if(argc < 1)
	{
		return 0;
	}
	
	int Sock = (int)lua_tonumber(L, 1);
	Disconnect(Sock);
	return 0;
}


void LuaSocketManager::Register(lua_State *L)
{
	lua_register(L, "Client", Lua_Connect_Wrap);
	lua_register(L, "Server", Lua_Server_Wrap);
	lua_register(L, "Accept", Lua_Accept_Wrap);
	lua_register(L, "Recv", Lua_Recv_Wrap);
	lua_register(L, "Send", Lua_Send_Wrap);
	lua_register(L, "DataAvaliable", Lua_DataAvaliable_Wrap);
	lua_register(L, "ClientWaiting", Lua_ClientWaiting_Wrap);
	lua_register(L, "Disconnect", Lua_Disconnect_Wrap);
}

void LuaSocketManager::Cleanup()
{
	for(int i = 0; i < 64; i++)
	{
		if(Sockets[i])
		{
			delete Sockets[i];
		}
	}
	memset((void *)Sockets, 0, 4*64);

	WSACleanup();
}


